// ======================================================================
//
// String.cpp
// Jason Hughes
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "first.h"
#include "VString.h"

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//  Static variables
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
int String::charsz = 0;
int String::numStringsAllocated = 0;

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//  Static functions
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

void String::install(int charsize)
{
	charsz=charsize;
	DEBUG_FATAL(numStringsAllocated,("There are strings currently allocated.  Cannot String::install().\n"));
}

void String::remove(void)
{
	charsz=0;
	DEBUG_FATAL(numStringsAllocated,("There are strings currently allocated.  Cannot String::remove().\n"));
}


int String::strlen(const void *str,int charsize)
{
	int index=0;

	switch (charsize)
	{
		case 1:
			return ::strlen(static_cast<const char *>(str));
		case 2:
			for (;; index++)
			{
				if (*(static_cast<const u_int16 *>(str)+index)==0)
					return index;
			}
		case 4:
			for (;; index++)
			{
				if (*(static_cast<const u_int32 *>(str)+index)==0)
					return index;
			}
		default:
			Fatal("String::charsize - invalid character size.\n");
			return 0;
	}
}

void String::strcpy(void *dest,const void *src,int srcLength,int charsize)
{
	if (srcLength==-1)
		srcLength=strlen(src,charsize);

	if (charsz!=charsize)
	{
		int loop;
		u_int32 temp=0;

		for (loop=0; loop<srcLength; loop++)
		{
			switch (charsize)  // source character size
			{
				case 1:
					temp=*(static_cast<const u_int8 *>(src)+loop);
					break;
				case 2:
					temp=*(static_cast<const u_int16 *>(src)+loop);
					break;
				case 4:
					temp=*(static_cast<const u_int32 *>(src)+loop);
					break;
			}
			switch (charsz)    // destination character size
			{
				case 1:
					*(static_cast<u_int8 *>(dest)+loop)=static_cast<u_int8>(temp);
					break;
				case 2:
					*(static_cast<u_int16 *>(dest)+loop)=static_cast<u_int16>(temp);
					break;
				case 4:
					*(static_cast<u_int32 *>(dest)+loop)=temp;
					break;
			}
		}
	}
	else memcpy(dest,src,srcLength*charsz);
	*(reinterpret_cast<u_int32 *>(static_cast<u_int8 *>(dest)+srcLength*charsz))=0;  // plop a 4-byte wide null there to cover all cases
}

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

String::String(const void *initString,int charsize)  // allocated buffer is EXACTLY the right size to hold this string, and no more
{
	DEBUG_FATAL(!charsz,("Install the string class before instantiating one.\n"));

	length=strlen(initString,charsize);
	allocSize=length*charsz+4;
	str=NEW u_int8[allocSize];
	strcpy(str,initString,length,charsize);

	numStringsAllocated++;
}

String::String(int allocsize)  // allocated buffer is whatever size designated, typically VS_DefaultLength
{
	DEBUG_FATAL(!charsz,("Install the string class before instantiating one.\n"));

	allocSize=allocsize*charsz;
	str=NEW u_int8[allocSize];
	length=0;
	setNull();

	numStringsAllocated++;
}

String::~String(void)
{
	delete []str;
	numStringsAllocated--;
}

void String::append(const void *appstr,int charsize)
{
	int appstrlen;

	appstrlen=strlen(appstr,charsize);
	if ((length+appstrlen)*charsz+4>=allocSize)  // allocate a space big enough to accept the append and delete the old space
	{
		u_int8 *newstr=NEW u_int8[(length+appstrlen)*charsz+4];
		memcpy(newstr,str,allocSize);
		allocSize=(length+appstrlen)*charsz+4;
		delete []str;
		str=newstr;
	}

	strcpy(str+length*charsz,appstr,appstrlen,charsize);
	length+=appstrlen;
}

void String::append(int appendChar)
{
	if (length*charsz+5>=allocSize)  // allocate more than enough space for more appendChars
	{
		u_int8 *newstr=NEW u_int8[length*charsz+25];
		memcpy(newstr,str,allocSize);
		allocSize=length*charsz+25;
		delete []str;
		str=newstr;
	}

	switch (charsz)    // destination character size
	{
		case 1:
			*(str+length)=static_cast<u_int8>(appendChar);
			break;
		case 2:
			*(reinterpret_cast<u_int16 *>(str)+length)=static_cast<u_int16>(appendChar);
			break;
		case 4:
			*(reinterpret_cast<u_int32 *>(str)+length)=appendChar;
			break;
	}
	length++;
	setNull(length);
}

int String::operator[](int index) const
{
	switch (charsz)
	{
		case 1:
			return *(str+index);
		case 2:
			return *(reinterpret_cast<const u_int16 *>(str)+index);
		case 4:
			return *(reinterpret_cast<const u_int32 *>(str)+index);
		default:
			Fatal("Cannot figure out how big the character size is to return from String::operator[]().\n");
			return 0;
	}
}

String &String::operator=(const String &src)
{
	length=0;
	setNull();
	append(src.str);
	return *this;
}

bool String::operator==(const String &src) const
{
	if (length!=src.length)
		return false;
	return (memcmp(str,src.str,length)==0);
}
